#!/usr/bin/env python3

__copyright__ = "Copyright (C) 2021-2022 Jesus M. Gonzalez-Barahona"
__license__ = "GPL-3.0-or-later"

import collections
import datetime
import hashlib
import json
import os
import re
import shelve
import sys

import magic
import ssdeep
import tlsh
import textdistance

this_dir = os.path.dirname(os.path.abspath(__file__))
scancode_cache_file = os.path.join(this_dir, 'scancode')
earliest_cache_file = os.path.join(this_dir, 'earliest')
swhid_cache_file = os.path.join(this_dir, 'swhid')


class Comp():

    def __init__(self, license):
        self.license = license

    attr_names = []

    def attrs(self):
        return self.attr_names

    def produce(self):
        if hasattr(self.license, self.attr_names[-1]):
            # We already produced, return
            return {}
        self.compute()
        a = {attr: getattr(self, attr) for attr in self.attrs()}
        return {attr: getattr(self, attr) for attr in self.attrs()}

class CompUnified(Comp):
    """Compute unified text, and related attributes
        * kind: mime type of text in file
        * codec: codec for text in file
        * text: text in file (as string)
        * subs_copyright: number of substitutions of the copyright pattern
        * removed_titles: number of titles (eg: "MIT License") removed
        * utext: unified text (as string, lowercase alphanumeric words
            separated by single spaces)
        * uhash: SHA1 hash of utext
    """
    attr_names = ['kind', 'codec', 'text', 'subs_copyright',
                  'removed_titles', 'utext', 'uhash']

    # To guess code and mime type for files
    guesser = magic.Magic(mime=True, mime_encoding=True)

    def compute(self):
        """Compute unified text for the license, if it was not computed earlier

        If kind is 'binary' all other attributes are None or 0.
        """
        with open(self.license.filename, 'rb') as f:
            content = f.read()
            kind, codec = self.guesser.from_buffer(content).split()
            self.codec = codec[8:]
            self.kind = kind[:-1]
            self.subs_copyright = 0
            self.removed_titles = 0
            if self.codec == 'binary':
                self.text = None
                self.utext = None
                self.uhash = None
            else:
                self.text, self.utext, self.codec, self.subs_copyright, self.removed_titles, self.uhash \
                    = self.compute_text(content, self.codec)

    @classmethod
    def compute_text(cls, content, codec):
        """Compute values for some content, corresponding to some text"""

        try:
            text = content.decode(codec, errors='ignore')
        except LookupError:
            codec = 'ascii'
            text = content.decode('ascii', errors='ignore')
        lines = []
        total_copyright = 0
        total_titles = 0
        # Deal with each line separately
        for line in text.splitlines():
            line = cls.normalize_chars(line)
            line, subs_copyright = cls.remove_copyright(line)
            total_copyright += subs_copyright
            line, removed_titles = cls.remove_titles(line)
            total_titles += removed_titles
            lines.append(line)
        # Now merge all lines and deal with them in a single string
        utext = ' '.join(lines)
        utext = cls.remove_repeated_spaces(utext)
        uhash = hashlib.sha1(utext.encode(codec)).hexdigest()
        return text, utext, codec, total_copyright, total_titles, uhash

    @classmethod
    def remove_repeated_spaces(cls, text):
        """Remove repeated spaces, substituting each occurrence for a single one

        Also, remove spaces at the start and at the end of text.

        :param text: text to explore
        :return: text with repeated spaces substituted
        """

        while '  ' in text:
            text = text.replace('  ', ' ')
        text = text.strip(' ')
        return text

    pa_norm_chars = re.compile(r'[^\w ]+')
    @classmethod
    def normalize_chars(cls, text):
        """Normalize characters, by removing everything which is not
        alphanumeric or space or comma, normalizing spaces to a single space,
        removing leading or trailing spaces, and lowercasing everything.

        :param text: text to explore
        :return: pair (text, occurrences)
        """
        result, occurrences = cls.pa_norm_chars.subn("", text)
        result = cls.remove_repeated_spaces(result)
        result = result.lower()
        return result

    # Regular expression for subs_copyright
    # Copyright line starts with like "copyright 2003 " or
    # "copyright c 2003 "
    # "All rights reserved" will also be removed
    pa_copyrights = [
        re.compile(r'^copyright( c)? \d{4}(\d{4})? .*$'), # "Copyright 2003 XXX", "Copyright (c) 2003-2005 XXX"
        re.compile(r'^copyright( c)? \d{4}(\d{4})?'), # "Copyright 2003", "Copyright (c) 2003-2005"
        re.compile(r'^all rights reserved*$') # All rights reserved
    ]
    @classmethod
    def remove_copyright(cls, text):
        """Remove lines which are copyright notices

         Copyright notices are considered to be of this form:
         * On a single line, maybe spaces or tabs at the beginning
         * "Copyright", "copyright" or "COPYRIGHT"
         * Then maybe "(C)"
         * Then a year (four digits), maybe repeated several times, separated by comma, space, -
         * Then characters up to the end of the line (usually the name (s))

        :param text: text to explore
        :return: pair (text, occurrences)
        """
        matches = 0
        for pa in cls.pa_copyrights:
            text, matched = pa.subn("", text)
            matches += matched
        return text, matches

    pa_titles = [
        re.compile(r'^license( mit)*$'), # License
        re.compile(r'^(the )*mit license( mit)*$'), # MIT
        re.compile(r'^bsd \dclause license$') # BSD
    ]
    @classmethod
    def remove_titles(cls, text):
        """Remove common titles such as "MIT License"
        """
        matches = 0
        for pa in cls.pa_titles:
            text, matched = pa.subn("", text)
            matches += matched
        return text, matches

    str_marks = '#*/'
    @classmethod
    def remove_marks(cls, text):
        """Remove marks such as comments (#, *, /,...)
        """
        result = text.translate(str.maketrans('', '', cls.str_marks))
        return result

class CompSSDeep(Comp):
    """Compute ssdeep hash
        * ssdeep: ssdeep hash of utext
    """

    attr_names = ['ssdeep']

    def compute(self):
        """Compute ssdeep hash of utext if codec is not 'binary'

        Needs compute_unified to be run earlier
        Attributes set: see attrs_names
        """
        if self.license.codec == 'binary':
            self.ssdeep = None
        else:
            self.ssdeep = ssdeep.hash(self.license.utext)


class CompTlsh(Comp):
    """Compute TLSH hash
        * tlsh: tlsh hash of utext or None if not computed
    """

    attr_names = ['tlsh']

    def compute(self):
        """Compute TLSH hash of utext

        TLSH will be set to None if codec is binary and/or if there was not
        enough content to compute a TLSH hash.

        Needs compute_unified to be run earlier

        Attributes set: see attrs_names

        """
        if self.license.codec == 'binary':
            self.tlsh = None
        else:
            # tlsh.hash() returns '' if not enough content is available; if so,
            # we convert it to None for consistency with other no-hash cases
            self.tlsh = tlsh.hash(self.license.utext.encode('utf8')) or None


class CompScanCode(Comp):
    """Read scancode result
        * scancode: scancode license id(s) (list)
        * scancode_p: scancode probability(ies) (list)
    """

    attr_names = ['scancode', 'scancode_p']

    cache_file = scancode_cache_file
    cache = None

    def compute(self):
        """Read scancode license

        Needs that read_scancode.py was run first, and that licenses were read
         in scancode_cache_file (a shelve file) earlier.
        """
        if self.cache == None:
            self.cache = shelve.open(self.cache_file)
        if self.license.name in self.cache:
            license_ids = [data['scancode'] for data in self.cache[self.license.name]]
            license_ps = [data['probability'] for data in self.cache[self.license.name]]
            self.scancode = ':'.join(license_ids)
            self.scancode_p = license_ps
        else:
            self.scancode = ''
            self.scancode_p = []


class CompIdentify(Comp):
    """Identify some kinds of file not interesting for analysis
        * valid: valid (interesting) for analusus (Boolean)
        * reason: reason (if invalid) (string)
    """

    attr_names = ['valid', 'reason']

    # Debian package license file
    # Example: 0000c1ccf7b1ceeb3a9ef72cd437d0c6dca6c7c8
    pa_debian_pkg = re.compile(r'this package was debianized by')
    # Debian package license file
    # Example: 00009b629c16e82219358cd163cdcca947b4f7c9
    pa_debian_pkg2 = re.compile(r'format httpswwwdebianorgdocpackagingmanualscopyrightformat')

    def compute(self):
        """Identify if valid, using several heuristics
        """
        if self.license.filesize < 150:
            self.valid = False
            self.reason = "too-short"
        elif self.license.utext is None:
            self.valid = False
            self.reason = "no-utext"
        elif self.license.kind != 'text/plain':
            self.valid = False
            self.reason = "no-textplain"
        elif (self.pa_debian_pkg.search(self.license.utext) is not None) or \
            (self.pa_debian_pkg2.search(self.license.utext) is not None):
            self.valid = False
            self.reason = "debian-pkg"
        else:
            self.valid = True
            self.reason = ""

class CompEarliest(Comp):
    """Find the earliest date for the file (license)
    Uses data from the earliest cache (database)
        * earliest: earliest date (datetime)
    """

    attr_names = ['earliest']

    earliest_file = earliest_cache_file
    earliest_cache = None
    swhid_file = swhid_cache_file
    swhid_cache = None

    def compute(self):
        """Read earliest date

        Needs that earliest dates were read in earliest_cache_file,
         and swhids were read in swhid_cache_cile (both shelve files) earlier.
        """
        if self.earliest_cache == None:
            self.earliest_cache = shelve.open(self.earliest_file)
        if self.swhid_cache == None:
            self.swhid_cache = shelve.open(self.swhid_file)
        try:
            swhid = self.swhid_cache[self.license.name]['swhid']
            self.earliest = datetime.datetime.fromtimestamp(self.earliest_cache[swhid]['date'])
        except KeyError:
            if self.license.name in self.swhid_cache:
                swhid = self.swhid_cache[self.license.name].get('swhid', None)
                if swhid is not None:
                    date = self.earliest_cache.get(swhid, None)
                else:
                    date = None
            else:
                swhid = None
                date = None
            # print("earliest not found:", self.license.name, swhid, date, file=sys.stderr)
            self.earliest = None


class CompWords(Comp):
    """Find out words in text
        * words: dictionary with list of words and count for each of them
    """

    # List of stopwords from NLTK
    stopwords = set(['ourselves', 'hers', 'between', 'yourself', 'but', 'again',
                 'there', 'about', 'once', 'during', 'out', 'very', 'having',
                 'with', 'they', 'own', 'an', 'be', 'some', 'for', 'do',
                 'its', 'yours', 'such', 'into', 'of', 'most', 'itself',
                 'other', 'off', 'is', 'am', 'or', 'who', 'as',
                 'from', 'him', 'each', 'the', 'themselves', 'until', 'below',
                 'are', 'we', 'these', 'your', 'his', 'through', 'dont', 'nor',
                 'me', 'were', 'her', 'more', 'himself', 'this', 'down',
                 'should', 'our', 'their', 'while', 'above', 'both', 'up',
                 'to', 'ours', 'had', 'she', 'all', 'no', 'when', 'at', 'any',
                 'before', 'them', 'same', 'and', 'been', 'have', 'in',
                 'will', 'on', 'does', 'yourselves', 'then', 'that', 'because',
                 'what', 'over', 'why', 'so', 'can', 'did', 'not', 'now',
                 'under', 'he', 'you', 'herself', 'has', 'just', 'where',
                 'too', 'only', 'myself', 'which', 'those', 'i', 'after', 'few',
                 'whom', 't', 'being', 'if', 'theirs', 'my', 'against', 'a',
                 'by', 'doing', 'it', 'how', 'further', 'was', 'here', 'than'])
    words_in_licenses = {'following', 'licensor', 'form', 'copyright', 'version', 'document', 'affirmer', 'original', 'library', 'derivative', 'general', 'free', 'provided', 'regents', 'works', 'use', 'program', 'related', 'license', 'standard', 'licence', 'covered', 'copy', 'code', 'shall', 'package', 'work', 'section', 'without', 'permission', 'may', 'software', 'gnu', 'contributor', 'licensee', 'source', 'rights'}
    stopwords_in_licenses = {'to', 'a', 'you', 'or', 'this', 'the', 'in', 'that', 'of', 'and'}

    attr_names = ['words', 'stopwords', 'license_words', 'license_stopwords']

    @classmethod
    def words_from_file(cls, words_file='licenses_words.json'):
        """Read words and stopwords from file

        :param words_file: file name with words and stopwords (JSON format)"""

        with open(words_file, 'r') as file:
            data = json.load(file)
            cls.words_in_licenses = set(data['words'])
            cls.stopwords_in_licenses = set(data['stopwords'])

    def compute(self):
        """Get list of words, and count for each of them, based on utext
        """
        if self.license.utext:
            all_words = self.license.utext.split()
            counter = collections.Counter(all_words)
            self.words = {word: counter[word] for word in counter.keys()
                          if word not in self.stopwords}
            self.stopwords = {word: counter[word] for word in counter.keys()
                              if word in self.stopwords}
            self.license_words = len(set(self.words.keys()) & self.words_in_licenses)
            self.license_stopwords = len(set(self.stopwords.keys()) & self.stopwords_in_licenses)
        else:
            self.words = []
            self.stopwords = []
            self.license_words = 0
            self.license_stopwords = 0

# Standard comps for License
std_comps = [CompUnified, CompSSDeep, CompTlsh, CompScanCode, CompIdentify]

class License():
    """Object with data for a license
    """

    @staticmethod
    def attrs_init():
        """List of attribute names set by __init__

        * name (unique name for the license file),
        * filename (path of the file with the text)
        * filesize (size of file)
        """
        return['name', 'filename', 'filesize']

    def __init__(self, filename, name, comps=std_comps):
        """Produce a License object, corresponding to the text in a file

        Attributes set: see attrs_init.

        :param filename: name of file to read (full path)
        :param name: name of the license
        """

        self.name = name
        self.filename = filename
        self.filesize = os.stat(filename).st_size
        for a_class in comps:
            comp = a_class(self)
            for attr, value in comp.produce().items():
                setattr(self, attr, value)

    def get_attrs(self, attr_list):
        "Get values for list of attributes, as a dictionary"
        return [getattr(self, attr) for attr in attr_list]

    def get_attrs_csv(self, attr_list):
        "Get values for list of attributes, as a csv string"
        return ','.join([str(attr) for attr in self.get_attrs(attr_list)])

