#!/usr/bin/env python3

__copyright__ = "Copyright (C) 2021 Jesus M. Gonzalez-Barahona"
__license__ = "GPL-3.0-or-later"

import atexit
import csv
import os
import pathlib
import shelve
import sys

import textdistance

import licen
from licen import License

def _get_files(dir):
    """Generator producing all files in a directory, at any level

    For each file found, a pair directory, filename is yielded
    """

    for path in sorted(os.listdir(dir)):
        full_path = os.path.join(dir, path)
        if os.path.isdir(full_path):
            for subdir, file in _get_files(full_path):
                yield subdir, file
        else:
            yield dir, path


class Licenses():
    """Class for a collection of licenses"""

    def __init__(self, paths=[], cache=None, rebuild_cache=False, comps=licen.std_comps):
        """Initializa object

        If paths correspond to files, they are considered as licenses.
        If they are directories, files in them, at any level, are considered
        as licenses.

        :param paths: list of paths with licenses to add to the collection
        :param cache: filename for the shelve cache for storing licenses
        :param rebuild_cache: rebuild the cache from scratch, if it exists
        :param comps: list of Comp classes to use for computing data of licenses
        """

        if not isinstance(paths, list):
            raise TypeError("Argument should be a list of paths")
        self.paths = paths
        if cache is None:
            self.cache = None
        else:
            if rebuild_cache:
                # Always create a new, empty database
                dbm_flag = 'n'
            else:
                # Open database, creating it if it doesn’t exist
                dbm_flag = 'c'
            self.cache = shelve.open(cache, flag=dbm_flag)
            atexit.register(self.close_cache)
        self.comps = comps

    def get_licenses(self):
        """Returns a generator with all files as License objects"""

        for path in self.paths:
            if os.path.isdir(path):
                for dir, file in _get_files(path):
                    license = self.get_license(os.path.join(dir, file), file)
                    yield license
            else:
                file = pathlib.Path(path).name
                license = self.get_license(path, file)
                yield license

    def get_license(self, file, name):
        if self.cache is None:
            license = License(file, name, comps=self.comps)
        else:
            try:
                license = self.cache[file]
            except KeyError:
                self.cache[name] = License(file, name, comps=self.comps)
                license = self.cache[name]
        return license

    def close_cache(self):
        self.cache.close()

    def len(self):
        return len(self.cache)

    def get_fields(self):
        """Get fields names for the comps that some instance of this class is using"""

        return [field for comp in self.comps for field in comp.attr_names]


class CachedLicenses():
    """Class for a collection of licenses, already stored in a cache"""

    def __init__(self, cache):
        """Initializa object

        :param cache: filename for the shelve cache for storing licenses
        """

        self.cache = shelve.open(cache, flag='r')
        atexit.register(self.close_cache)

    def get_licenses(self):
        """Returns a generator with all licenses as License objects"""

        for name in self.cache:
            license = self.cache[name]
            yield license

    def get_license(self, name):
        license = self.cache[name]
        return license

    def close_cache(self):
        self.cache.close()

    def len(self):
        return len(self.cache)


class CanonLicenses():
    """Read list of canonical licenses from a list of files in a directory

    Produce a data structure with a list of licenses.
    """

    def __init__(self, dir, files):
        """Read licenses from Debian licenses directory"""
        self.licenses = []
        for file in files:
            license = License(os.path.join(dir, file), file)
            self.licenses.append(license)

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < len(self.licenses):
            self.n += 1
            return(self.licenses[self.n - 1])
        else:
            raise StopIteration

class CanonDebianLicenses(CanonLicenses):

    files = [
        "Apache-2.0", "CC0-1.0", "GFDL-1.2", "GPL-1", "GPL-3",
        "LGPL-2.1", "MPL-2.0", "Artistic", "eupl-1.1.txt",
        "GFDL-1.3", "GPL-2", "LGPL-3", "BSD", "GFDL",
        "gpl-2.0.txt", "LGPL-2", "MPL-1.1"
    ]
    dir = "/usr/share/common-licenses"

    def __init__(self):
        super().__init__(self.dir, self.files)

class CanonSelectedLicenses(Licenses):

    dir = 'licenses/canon'
    def __init__(self, *args, **kwargs):
        super().__init__([self.dir], *args, **kwargs)

class OSISelectedLicenses(Licenses):

    dir = 'licenses/osi'
    def __init__(self, *args, **kwargs):
        super().__init__([self.dir], *args, **kwargs)
        super().__init__([self.dir], *args, **kwargs)

class ScancodeSelectedLicenses(Licenses):

    dir = 'licenses/scancode'
    def __init__(self, *args, **kwargs):
        super().__init__([self.dir], *args, **kwargs)
        super().__init__([self.dir], *args, **kwargs)

class ComparingLicenses():
    """Class for comparing licenses"""

    def __init__(self, canon, dir):
        """Initializa object

        :param canon: object with canonical list of licenses (to compare with)
        :param dir: directory with licenses to compare
        """

        self.canon = canon
        self.dir = dir

    def _get_files(self, dir=None):
        if dir is None:
            dir = self.dir
        for path in sorted(os.listdir(dir)):
            full_path = os.path.join(dir, path)
            if os.path.isdir(full_path):
                for subdir, file in self._get_files(full_path):
                    yield subdir, file
            else:
                yield dir, path

    def get_licenses(self):
        """Returns a generator with all files as License objects"""
        for path, file in self._get_files():
            license = License(os.path.join(path, file), file)
            yield license

    def compare_file(self, path, name, algo, output=sys.stdout):
        license = License(path, name)
        if license.text is not None:
            metrics = {}
            for canon_license in self.canon:
                metrics[canon_license.name] = algo(canon_license.text, license.text)
            return {**{'name': license.name}, **license.dict(), **metrics}
        else:
            return None

    def compare(self, algo, output=sys.stdout):
        """Compare canonical licenses with licenses in directory

        Produces a CSV file with a comparison of canonical licenses with
        all licenses in directory dir (one license per file)
        :param dir: directory with all license files to compare with
        """
        license_names = [license.name for license in self.canon]
        header_names = ['name'] + License.fields() + license_names
        writer = csv.DictWriter(output, fieldnames=header_names)
        writer.writeheader()
        for path, file in self._get_files():
            metrics = self.compare_file(os.path.join(path, file), file, algo)
            if metrics is not None:
                writer.writerow(metrics)


if __name__ == '__main__':
    #canon_licenses = CanonDebianLicenses()
    canon_licenses = CanonSelectedLicenses()

    #licenses = Licenses(canon_licenses, 'data')
    licenses = Licenses(canon_licenses, '../sh-licenses/00/00')

    #licenses.compare(textdistance.hamming.normalized_similarity)
    #licenses.compare(textdistance.entropy_ncd.normalized_similarity)
    with open('licenses_bag.csv', 'w') as output:
        licenses.compare(textdistance.bag.normalized_similarity, output)
