#!/usr/bin/env python3

__copyright__ = "Copyright (C) 2021 Jesus M. Gonzalez-Barahona"
__license__ = "GPL-3.0-or-later"

"""Produce a CSV with a list of files, to be manually analyzed as ground truth

The list of files produced is based on a list of license names, obtained from
a file with one license name per line.

Read file:
 * truth/filenames-random-{seed}.txt
 * truth/scancode-random-{seed}.csv
 
Produces two files:
 * truth/to_validate-{seed}.csv: all files to validate, their uhash, and the rest of
 fields to produce truth.csv
 * truth/to_validate_uhash-{seed}.csv only with one representant per uhash
 
Also produces a cache for scancode data:
  * truth/scancode-random-{seed}[.db]

{seed} is the seed used to generate the random list of files.

This script does not use any previous cache (it produces, and uses,
 its own, for scancode data).
"""

import argparse
import csv
import os
import shelve

import config
import licen
import licens

truth_dir = 'truth'
licenses_dir = config.all_licenses_dir

def parse_args():
    """Parse arguments, and return args with those found"""
    parser = argparse.ArgumentParser(description='Produce CSV file with list of files to annotate manually.')
    parser.add_argument('--seed', type=int, default=1,
                        help="seed used for filenames (default: 1)")
    return parser.parse_args()

def read_scancode(scancode_file, scancode_cache):
    """Read a CSV with scancode data for files, producing a cache suitable for CompScanCode

    Assumes CSV file is in the format provided by the SWH dataset.
    """

    # 'n'  flag: create a new, empty database
    with shelve.open(scancode_cache, flag='n') as cache:
        with open(scancode_file, 'r') as csvfile:
            reader = csv.reader(csvfile)
            for row in reader:
                if row[0].lower() == 'sha1':  # CSV header line
                    continue
                file, identified, probability = row
                probability = float(probability)
                data = {'scancode': identified, 'probability': probability}
                if file in cache:
                    # File already in cache, add new license to it
                    data_list = cache[file]
                    data_list.append(data)
                else:
                    data_list = [data]
                cache[file] = data_list

def get_paths(filenames_file, licenses_dir=licenses_dir):
    """Get paths to licenses from file with filenames (one per line)"""

    def path_from_name(name, licenses_dir=licenses_dir):
        """Compute the path to a file, having its name

        Assumption: all files are in a two-level subdirectory,
        following conventions in the SWH dataset.
        """

        path = os.path.join(licenses_dir, name[0:2], name[2:4], name)
        return path

    with open(filenames_file) as file:
        paths = [path_from_name(row.strip()) for row in file]
    return paths

def validation_files(licenses, hdrs, filename_to_validate, filename_to_validate_uhash):
    """Produce list of files to validate manually, as CSV files

    Produces two lists, each in a CSV file, with same data for each file,
    but the second one grouped by uhash (a single representant for each uhash
    group. Data is produced by automatically annotating each file.
    """

    def prefill (license):

        if license.kind != 'text/plain':
            validation = [False, False, False, True, False, False, False, False, "Not text/plain"]
            selse = True
        elif license.filesize < 30:
            validation = [False, False, False, True, False, False, False, False, "Shorter than 30 chars"]
            selse = True
        else:
            validation = [False, False, False, False, False, False, False, False, ""]
            selse = False

        return {'validation': validation, 'selse': selse}

    uhashes = set()
    count = 0
    count_selse = 0
    count_licen = 0

    with open(filename_to_validate, 'w') as to_validate, \
         open(filename_to_validate_uhash, 'w') as to_validate_uhash:
        print(','.join(hdrs), file=to_validate)
        print(','.join(hdrs), file=to_validate_uhash)

        for license in licenses.get_licenses():
            count += 1
            license_data = license.get_attrs_csv(attrs)
            prefilled = prefill(license)
            license_data = license_data + ',' + ','.join([str(i) for i in prefilled['validation']])
            if prefilled['selse']:
                count_selse += 1
            else:
                count_licen += 1
            print(license_data, file=to_validate)
            if license.uhash not in uhashes:
                uhashes.add(license.uhash)
                print(license_data, file=to_validate_uhash)

    return{'total': count, 'uhashes': len(uhashes),
           'licen': count_licen, 'selse': count_selse}

if __name__ == '__main__':

    args = parse_args()
    seed = args.seed
    scancode_cache = os.path.join(truth_dir, f'scancode-random-{seed}')
    scancode_file = os.path.join(truth_dir, f'scancode-random-{seed}.csv')
    filenames_file = os.path.join(truth_dir, f'filenames-random-{seed}.txt')
    filename_to_validate = os.path.join(truth_dir, f'to_validate-{seed}.csv')
    filename_to_validate_uhash = os.path.join(truth_dir, f'to_validate_uhash-{seed}.csv')

    read_scancode(scancode_file=scancode_file, scancode_cache=scancode_cache)
    paths = get_paths(filenames_file=filenames_file, licenses_dir=licenses_dir)

    # Cache file for scancode results for each file
    licen.CompScanCode.cache_file = scancode_cache
    # Licenses object with all licenses in a directory, and the intended Comp classes
    licenses = licens.Licenses(paths,
                               cache=None, rebuild_cache = True,
                               comps=[licen.CompUnified, licen.CompScanCode, licen.CompIdentify])
    # Attributes to obtain for each license
    attrs = ['name', 'uhash', 'filesize', 'codec', 'kind', 'scancode',]
    # Header names to print in the CSV file (same order as attrs)
    hdrs = ['name', 'uhash', 'length', 'codec', 'mime', 'scancode']
    validation_hdrs = ['licen', 'notice', 'multi', 'selse', 'copy', 'swrong', 'debian', 'done', 'comment']
    hdrs = hdrs + validation_hdrs

    counts = validation_files(licenses, hdrs, filename_to_validate, filename_to_validate_uhash)

    print(f"Total files: {counts['total']}, total uhash files: {counts['uhashes']}.", end='')
    print(f" Maybe licen info: {counts['licen']}, likely something else: {counts['selse']}")