#!/usr/bin/env python3

__copyright__ = "Copyright (C) 2021 Jesus M. Gonzalez-Barahona"
__license__ = "GPL-3.0-or-later"

"""Train and analyze a Random Forest model to predict which documents are licenses

Deals only with plain text documents, and classifies using the "licen" field
in the Annotated Sample.
"""

import os, sys
from pathlib import Path
sys.path.append(str(Path.cwd().parent))

import pandas as pd
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.metrics import classification_report, confusion_matrix, accuracy_score
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier

import licens
import licen
import config

# Directory with all licenses, in subdirs of the form 00/00
licenses_dir = os.path.join('..', config.all_licenses_dir)
# File with the manually annotated ground truth (one line per file)
truth_file = os.path.join('truth.csv')

def path_from_name(name, licenses_dir=licenses_dir):
    """Compute the path to a file, having its name

    Assumption: all files are in a two-level subdirectory,
    following conventions in the SWH dataset.
    """

    path = os.path.join(licenses_dir, name[0:2], name[2:4], name)
    return path

# Read the Annotated Sample into a dataframe
truth_df = pd.read_csv(truth_file)

# Get all documents in the sample, and compute their normalized version
paths = [path_from_name(name, licenses_dir)
         for name in truth_df['name']]
documents = licens.Licenses(paths,
        cache=None, rebuild_cache = True,
        comps=[licen.CompUnified, licen.CompIdentify])

# Get a list with the normalized text for all plain text documents
text_licen = [licen
              for licen in documents.get_licenses()
              if (licen.utext is not None) and (licen.kind == 'text/plain')]
udocuments = [documents.get_license(licen.filename, licen.name).utext
             for licen in text_licen]

# Source values: Vectorized normalized documents (TF-IDF representation)
vectorizer = CountVectorizer(max_features=1500, min_df=5, max_df=0.7)
X = vectorizer.fit_transform(udocuments).toarray()
tfidfconverter = TfidfTransformer()
X = tfidfconverter.fit_transform(X).toarray()

# Target values: True if document includes a singl full-text license
y = [truth_df.loc[truth_df['name'] == doc.name, 'licen'].values[0]
     for doc in text_licen]

# Get training and testing samples
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=0)

# Train a Random Forest classifier
classifier = RandomForestClassifier(n_estimators=1000, random_state=0, n_jobs=-1)
classifier.fit(X_train, y_train)

# Predict which files are full-text licenses for testing sample
y_pred = classifier.predict(X_test)

# Print results of the prediction
print(confusion_matrix(y_test,y_pred))
print(classification_report(y_test,y_pred))
print(accuracy_score(y_test, y_pred))